/*

convolve~

Copyright 2010 William Brent

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <http://www.gnu.org/licenses/>.

version 0.9, Dec 16, 2010

 0.9 was shifting one block too few in nonoverlapped_output. this fixes it
 0.8 deals with x->num_parts+1 issue
 0.7 cleanup version of 0.6, this one is possibly ready for use. next, should do a #define for a maximum number of partitions. should also deal with (x->num_parts+1) in the free routine.  it's a special case if we free with x->num_parts = 0...otherwise we try to free too much memory.

*/

#include "m_pd.h"
#include <math.h>
#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif
#define MINWIN 256

static t_class *convolve_tilde_class;

typedef struct _convolve_tilde 
{
    t_object x_obj;
    t_symbol *arrayname;
    t_word *x_vec;
    int arraysize;
    int num_parts;
    float sr;
    float n;
    int dsp_tick;
    int buffer_limit;
    int window;
    int window_double;
    float amp_scalar;
    t_sample *ir_spec;
	t_sample *signal_buf;
	t_sample *signal_buf_padded;
	
	t_sample *nonoverlapped_output;
	t_sample *final_output;
	
    float x_f;
    
} convolve_tilde;


/* ------------------------ convolve~ -------------------------------- */

static void convolve_tilde_analyze(convolve_tilde *x, t_symbol *s)
{
    t_garray *a;
    int i, j, oldsize, newsize, oldsize_plus1, newsize_plus1;

	oldsize = x->num_parts*x->window_double;
	
	// if the current number of parts is zero, oldsize_plus1 should also be zero
	if(x->num_parts)
		oldsize_plus1 = (x->num_parts+1)*x->window_double;
	else
		oldsize_plus1 = 0;

	x->arrayname = s;
    
    if (!(a = (t_garray *)pd_findbyclass(x->arrayname, garray_class)))
    {
        if (*s->s_name) pd_error(x, "convolve~: %s: no such array",
            x->arrayname->s_name);
        x->x_vec = 0;
    }
    else if (!garray_getfloatwords(a, &x->arraysize, &x->x_vec))
    {
        pd_error(x, "%s: bad template for convolve~", x->arrayname->s_name);
        x->x_vec = 0;
    }
    else garray_usedindsp(a);
    
    x->num_parts = 0;
    while((x->num_parts*x->window) < x->arraysize)
    	x->num_parts++;
    
    newsize = x->num_parts*x->window_double;
    newsize_plus1 = (x->num_parts+1)*x->window_double;
	
    x->ir_spec = (t_sample *)t_resizebytes(x->ir_spec, oldsize*sizeof(t_sample), newsize*sizeof(t_sample));
    x->nonoverlapped_output = (t_sample *)t_resizebytes(x->nonoverlapped_output, oldsize_plus1*sizeof(t_sample), newsize_plus1*sizeof(t_sample));

    // copy array internally for upcoming in-place FFT
    for(i=0; i<x->num_parts; i++)
    {
    	int start_spec, start_vec;
    	
    	start_spec = i*x->window_double;
    	start_vec = i*x->window;
    	
		for(j=0; j<x->window && (start_vec+j)<x->arraysize; j++)
			x->ir_spec[start_spec+j] = x->x_vec[start_vec+j].w_float;
		
		// zero pad a gap before the next window
		for(; j<x->window_double; j++)
			x->ir_spec[start_spec+j] = 0.0;
	}

    for(i=0; i<newsize_plus1; i++)
    	x->nonoverlapped_output[i] = 0.0;
    
    // take FTs of IR partitions, not Hann windowed
	for(i=0; i<x->num_parts; i++)
	{
		int idx;
		
		idx = i*x->window_double;

		mayer_realfft(x->window_double, x->ir_spec+idx);
	}
	
	post("analysis complete: %i partitions", x->num_parts);
}


static void *convolve_tilde_new(t_floatarg window)
{
    convolve_tilde *x = (convolve_tilde *)pd_new(convolve_tilde_class);
	int i, isPow2;
	
	outlet_new(&x->x_obj, &s_signal);

	if(window)
	{
		isPow2 = (int)window && !( ((int)window-1) & (int)window );
		if(isPow2)
		{
			if(window >= MINWIN)
				x->window = window;
			else
			{
				x->window = MINWIN;
				error("convolve~: requested window size too small. minimum value of %i used instead", x->window);				
			}
		}
		else
		{
			x->window = 4096;
			error("convolve~: window not a power of 2. default value of %i used instead", x->window);
		}
	}
	else
	{
		x->window = 4096;
	}
	
	x->arraysize = 0;
	x->num_parts = 0;
	x->sr = 44100;
	x->n = 64;
	x->dsp_tick = 0; 
	x->window_double = x->window*2;
	x->amp_scalar = 1.0f/x->window_double;
	
	x->buffer_limit = x->window/x->n;

	// these will be resized when analysis occurs
	x->ir_spec = (t_sample *)getbytes(0);
	x->nonoverlapped_output = (t_sample *)getbytes(0);
	
	x->signal_buf = (t_sample *)getbytes(x->window*sizeof(t_sample));
	x->signal_buf_padded = (t_sample *)getbytes(x->window_double*sizeof(t_sample));
	x->final_output = (t_sample *)getbytes(x->window*sizeof(t_sample));


	// init signal buffers
 	for(i=0; i<x->window; i++)
 		x->signal_buf[i] = 0.0;

 	for(i=0; i<x->window_double; i++)
 		x->signal_buf_padded[i] = 0.0;

	// init output buffer
	for(i=0; i<x->window; i++)
		x->final_output[i] = 0.0;
	
    post("convolve~: partition size %i", x->window);
    
    return (x);
}


static t_int *convolve_tilde_perform(t_int *w)
{
    int i, j, p, n, window, window_double, num_parts;
	float amp_scalar;
	
    convolve_tilde *x = (convolve_tilde *)(w[1]);

    t_sample *in = (t_float *)(w[2]);
    t_sample *out = (t_float *)(w[3]);
    
    n = w[4];
	window = x->window;
	window_double = x->window_double;
	num_parts = x->num_parts;
	amp_scalar = x->amp_scalar;
		
	// buffer most recent block
	for(i=0; i<n; i++)
		x->signal_buf[(x->dsp_tick*n)+i] = in[i];
		
	if(++x->dsp_tick >= x->buffer_limit)
	{
		x->dsp_tick = 0;
		
		// don't do anything if the IR hasn't been analyzed yet
		if(x->num_parts)
		{
		// copy the signal buffer into the in-place transform buffer
 		for(i=0; i<window; i++)
 			x->signal_buf_padded[i] = x->signal_buf[i];
		
		// pad the rest out with zeros
		for(; i<window_double; i++)
			x->signal_buf_padded[i] = 0.0;

		// take FT of the input window
		mayer_realfft(window_double, x->signal_buf_padded);

 		// multiply against partitioned IR spectra. these need to be complex multiplies
 		// also, sum into appropriate part of the nonoverlapped buffer
		for(p=0; p<num_parts; p++)
		{
			int ir_start, output_start;
			
			ir_start = p*window_double;
			output_start = (p+1)*window_double; // start summing into the second chunk (1) of the nonoverlapped output. the first chunk (0) is the time domain previous block, whose second half will be summed with the first half of this second chunk after the IFT below.
			
			// DC and Nyquist are real only, hence normal multiplies
			x->nonoverlapped_output[output_start] += x->signal_buf_padded[0] * x->ir_spec[ir_start];
			
			x->nonoverlapped_output[output_start+window] += x->signal_buf_padded[window] * x->ir_spec[ir_start+window];
			
			for(i=1, j=window_double-1; i<window; i++, j--)
			{
				float real, imag;
				
				// MINUS the imag part because i^2 = -1
				real = (x->signal_buf_padded[i] * x->ir_spec[ir_start+i]) - (x->signal_buf_padded[j] * x->ir_spec[ir_start+j]);
				
				imag = (x->signal_buf_padded[i] * x->ir_spec[ir_start+j]) + (x->signal_buf_padded[j] * x->ir_spec[ir_start+i]);
				
				x->nonoverlapped_output[output_start+i] += real;
				x->nonoverlapped_output[output_start+j] += imag;
			}
		}
		
		// resynth just the second x->window_double block of points in the output buffer. the first block is the previous output
		mayer_realifft(window_double, x->nonoverlapped_output+window_double);
	
		// write time domain output to x->final_output and reduce gain
		for(i=0; i<window; i++)
		{
			x->final_output[i] = x->nonoverlapped_output[window+i] +  x->nonoverlapped_output[window_double+i];

			x->final_output[i] *= amp_scalar;
		}
		
		// push remaining output buffer contents (all but one are still in freq domain) backwards
		for(i=0; i<((x->num_parts+1)*x->window_double)-window_double; i++)
			x->nonoverlapped_output[i] = x->nonoverlapped_output[window_double+i];
		
		// init the newly available chunk at the end
		for(; i<(x->num_parts+1)*x->window_double; i++)
			x->nonoverlapped_output[i] = 0.0;

 		}
 
	};
	
	// output
	for(i=0; i<n; i++, out++)
		if(x->num_parts)
			*out = x->final_output[(x->dsp_tick*n)+i];
		else
			*out = 0.0;

    return (w+5);
}


static void convolve_tilde_dsp(convolve_tilde *x, t_signal **sp)
{
	dsp_add(
		convolve_tilde_perform,
		4,
		x,
		sp[0]->s_vec,
		sp[1]->s_vec,
		sp[0]->s_n
	);

	if(!x->num_parts)
		error("convolve~: impulse response analysis not performed yet. output will be zero.");
};

static void convolve_tilde_free(convolve_tilde *x)
{	
    t_freebytes(x->ir_spec, x->num_parts*x->window_double*sizeof(t_sample));
    t_freebytes(x->signal_buf, x->window*sizeof(t_sample));
    t_freebytes(x->signal_buf_padded, x->window_double*sizeof(t_sample));

    if(x->num_parts)
	    t_freebytes(x->nonoverlapped_output, (x->num_parts+1)*x->window_double*sizeof(t_sample));

    t_freebytes(x->final_output, x->window*sizeof(t_sample));
};

void convolve_tilde_setup(void)
{
    convolve_tilde_class = 
    class_new(
    	gensym("convolve~"),
    	(t_newmethod)convolve_tilde_new,
    	(t_method)convolve_tilde_free,
        sizeof(convolve_tilde),
        CLASS_DEFAULT,
        A_DEFFLOAT,
		0
    );

    CLASS_MAINSIGNALIN(convolve_tilde_class, convolve_tilde, x_f);
	
	class_addmethod(
		convolve_tilde_class,
		(t_method)convolve_tilde_analyze,
		gensym("analyze"),
		A_SYMBOL,
		0
	);
	
    class_addmethod(
    	convolve_tilde_class,
    	(t_method)convolve_tilde_dsp,
    	gensym("dsp"),
    	0
    );
}
